@@ -5,7 +5,7 @@ import astroid
 from pylint.checkers import BaseChecker, utils
 from pylint.interfaces import IAstroidChecker
 
-from .transform import transform_module, is_some_protobuf_module, to_pytype, is_composite, is_repeated
+from .transform import transform_module, is_some_protobuf_module, to_pytype, is_composite, is_repeated, is_oneof
 from .transform import SimpleDescriptor, PROTOBUF_IMPLICIT_ATTRS, PROTOBUF_ENUM_IMPLICIT_ATTRS
 
 _MISSING_IMPORT_IS_ERROR = False
@@ -13,10 +13,10 @@ BASE_ID = 59
 MESSAGES = {
     'E%02d01' % BASE_ID: (
         'Field %r does not appear in the declared fields of protobuf-'
-        'generated class %r and will raise AttributeError on access',
+        'generated class %r',
         'protobuf-undefined-attribute',
         'Used when an undefined field of a protobuf generated class is '
-        'accessed'
+        'accessed, assigned, or passed to HasField/ClearField'
     ),
     'E%02d02' % BASE_ID: (
         'Value %r does not appear in the declared values of protobuf-'
@@ -44,32 +44,32 @@ MESSAGES = {
         'Used when a value is written to a read-only field such as a '
         'composite or repeated field'
     ),
+    'E%02d06' % BASE_ID: (
+        'Repeated fields cannot be checked for membership',
+        'protobuf-no-repeated-membership',
+        'Used when calling HasField/ClearField on a repeated field'
+    ),
+    'E%02d07' % BASE_ID: (
+        'Non-optional, non-submessage field %r cannot be checked for '
+        'membership in proto3 messages',
+        'protobuf-no-proto3-membership',
+        'Used when calling HasField/ClearField on a non-composite, '
+        'non-oneof field in a proto3 syntax file'
+    ),
+    'E%02d08' % BASE_ID: (
+        'Extension field %r does not belong to message %r',
+        'protobuf-wrong-extension-scope',
+        'Used when attempting to set an extension on a message when '
+        'it is not the containing type for that field'
+    ),
 }
-WELLKNOWNTYPE_MODULES = [
-    'any_pb2',
-    'descriptor_pb2',
-    'duration_pb2',
-    'empty_pb2',
-    'field_mask_pb2',
-    'struct_pb2',
-    'timestamp_pb2',
-    'type_pb2',
-    'wrappers_pb2',
-]
 Node = astroid.node_classes.NodeNG
 
 
-def wellknowntype(node):
-    # type: (astroid.Instance) -> bool
-    modname, type_ = node.pytype().rsplit('.', 1)
-    return (modname in WELLKNOWNTYPE_MODULES) or \
-           (modname.startswith('google.protobuf') and modname.endswith('_pb2'))
-
-
 def _get_inferred_values(node):
     # type: (Node) -> List[Node]
     try:
-        vals = node.inferred()
+        vals = list(node.infer())  # not the same as node.inferred() for astroid.Instance
     except astroid.InferenceError:
         return []
     return [v for v in vals if v is not astroid.Uninferable]
@@ -156,6 +156,8 @@ class ProtobufDescriptorChecker(BaseChecker):
         self._check_init_posargs(node)
         self._check_init_kwargs(node)
         self._check_repeated_scalar(node)
+        self._check_repeated_composite(node)
+        self._check_hasfield(node)
 
     @utils.check_messages('protobuf-enum-value')
     def _check_enum_values(self, node):
@@ -182,71 +184,116 @@ class ProtobufDescriptorChecker(BaseChecker):
                 self.add_message('protobuf-enum-value', args=(val, desc.name), node=node)
                 break  # should we continue to check?
 
-    @utils.check_messages('protobuf-no-posargs')
     def _check_init_posargs(self, node):
         # type: (astroid.Call) -> None
         desc = _get_protobuf_descriptor(node.func)
-        if desc is not None and len(node.args) > 0:
-            self.add_message('protobuf-no-posargs', node=node)
+        self._check_posargs(desc, node)
 
-    @utils.check_messages('protobuf-type-error')
     def _check_init_kwargs(self, node):
         # type: (astroid.Call) -> None
         desc = _get_protobuf_descriptor(node.func)
+        self._check_kwargs(desc, node)
+
+    def _check_repeated_composite(self, node):
+        # type: (astroid.Call) -> None
+        func = node.func
+        if not isinstance(func, astroid.Attribute) or func.attrname != 'add':
+            return # Call should look like "message.inner.add(**kwargs)"
+        desc = None  # type: Optional[SimpleDescriptor]
+        for val in _get_inferred_values(node):
+            desc = _get_protobuf_descriptor(val)
+            if desc is not None:
+                break  # I don't think there's a reasonable solution for ambiguous cases
+        self._check_posargs(desc, node)
+        self._check_kwargs(desc, node)
+
+    @utils.check_messages('protobuf-no-posargs')
+    def _check_posargs(self, desc, node):
+        # type: (Optional[SimpleDescriptor], astroid.Call) -> None
+        if desc is not None and len(node.args) > 0:
+            self.add_message('protobuf-no-posargs', node=node)
+
+    @utils.check_messages('protobuf-type-error', 'unexpected-keyword-arg')
+    def _check_kwargs(self, desc, node):
+        # type: (Optional[SimpleDescriptor], astroid.Call) -> None
         if desc is None:
             return
+        desc_name = desc.name
         keywords = node.keywords or []
         for kw in keywords:
             arg_name, val_node = kw.arg, kw.value
             if arg_name not in desc.fields_by_name:
-                continue  # should raise "unexpected-keyword-arg"
+                # This is probably too fragile (should it be replaced by a separate message?)
+                # Also, it's probably not fair to always refer to this as a constructor call, though arguably
+                # .add() on repeated composite fields is "constructing" the sub-message
+                self.add_message('unexpected-keyword-arg', node=node, args=(arg_name, 'constructor'))
+                continue
             fd = desc.fields_by_name[arg_name]
             arg_type = to_pytype(fd)
-            if is_composite(fd):
-                # messages
-                for val in _get_inferred_values(val_node):
-                    if isinstance(val, astroid.Const):
-                        if val.value is not None:
-                            # Special-case None as default of keyword args
-                            self.add_message('protobuf-type-error', node=node,
-                                             args=(desc.name, arg_name, arg_type.__name__, val.value))
-                        break
-                    val_desc = _get_protobuf_descriptor(val)
-                    if val_desc is None:
-                        continue  # XXX: ignore?
-                    if not val_desc.is_typeof_field(fd):
-                        val = '{}()'.format(val_desc.name)
-                        self.add_message('protobuf-type-error', node=node,
-                                         args=(desc.name, arg_name, arg_type.__name__, val))
-            else:
-                for val_const in _get_inferred_values(val_node):
-                    if hasattr(val_const, 'value'):
-                        val = val_const.value
-                    elif hasattr(val_const, '_proxied'):
-                        # check for <Instance of builtins.int> etc.
-                        val = _resolve_builtin(val_const)
-                        if val is None:  # check for composite assigned to scalar
-                            try:
-                                val = '{}()'.format(val_const.name)
-                            except AttributeError:
-                                continue
-                            self.add_message('protobuf-type-error', node=node,
-                                             args=(desc.name, arg_name, arg_type.__name__, val))
-                            break
+            if isinstance(val_node, astroid.Call):
+                val_node = _get_inferred_values(val_node)
+                if len(val_node) != 1:
+                    continue  # don't reason about ambiguous cases or uninferable
+                val_node = val_node[0]
+            if is_repeated(fd):
+                # What to do about iterators? At a certain point, inference becomes too complicated
+                elements = getattr(val_node, 'elts', [])
+                for element in elements:
+                    if is_composite(fd):
+                        self._check_composite_keyword(arg_name, arg_type, fd, node, element, desc_name)
                     else:
-                        try:
-                            val = '{}()'.format(val_const.name)
-                        except AttributeError:
-                            continue
-                        self.add_message('protobuf-type-error', node=node,
-                                         args=(desc.name, arg_name, arg_type.__name__, val))
-                        break
-                    if not _scalar_typecheck(val, arg_type):
-                        if val is not None:
-                            # Special-case None as default of keyword args
-                            self.add_message('protobuf-type-error', node=node,
-                                             args=(desc.name, arg_name, arg_type.__name__, val))
-                        break
+                        self._check_noncomposite_keyword(arg_name, arg_type, node, element, desc_name)
+            else:
+                if is_composite(fd):
+                    self._check_composite_keyword(arg_name, arg_type, fd, node, val_node, desc_name)
+                else:
+                    self._check_noncomposite_keyword(arg_name, arg_type, node, val_node, desc_name)
+
+    def _check_composite_keyword(self, arg_name, arg_type, fd, node, val_node, desc_name):
+        for val in _get_inferred_values(val_node):
+            if isinstance(val, astroid.Const):
+                if val.value is not None:
+                    # Special-case None as default of keyword args
+                    self.add_message('protobuf-type-error', node=node,
+                                     args=(desc_name, arg_name, arg_type.__name__, val.value))
+                break
+            val_desc = _get_protobuf_descriptor(val)
+            if val_desc is None:
+                continue  # XXX: ignore?
+            if not val_desc.is_typeof_field(fd):
+                val = '{}()'.format(val_desc.name)
+                self.add_message('protobuf-type-error', node=node,
+                                 args=(desc_name, arg_name, arg_type.__name__, val))
+
+    def _check_noncomposite_keyword(self, arg_name, arg_type, node, val_node, desc_name):
+        for val_const in _get_inferred_values(val_node):
+            if hasattr(val_const, 'value'):
+                val = val_const.value
+            elif hasattr(val_const, '_proxied'):
+                # check for <Instance of builtins.int> etc.
+                val = _resolve_builtin(val_const)
+                if val is None:  # check for composite assigned to scalar
+                    try:
+                        val = '{}()'.format(val_const.name)
+                    except AttributeError:
+                        continue
+                    self.add_message('protobuf-type-error', node=node,
+                                     args=(desc_name, arg_name, arg_type.__name__, val))
+                    break
+            else:
+                try:
+                    val = '{}()'.format(val_const.name)
+                except AttributeError:
+                    continue
+                self.add_message('protobuf-type-error', node=node,
+                                 args=(desc_name, arg_name, arg_type.__name__, val))
+                break
+            if not _scalar_typecheck(val, arg_type):
+                if val is not None:
+                    # Special-case None as default of keyword args
+                    self.add_message('protobuf-type-error', node=node,
+                                     args=(desc_name, arg_name, arg_type.__name__, val))
+                break
 
     @utils.check_messages('protobuf-type-error')
     def _check_repeated_scalar(self, node):
@@ -305,6 +352,37 @@ class ProtobufDescriptorChecker(BaseChecker):
         for val in vals:
             check_arg(val)
 
+    @utils.check_messages(
+        'protobuf-undefined-attribute',
+        'protobuf-no-repeated-membership',
+        'protobuf-no-proto3-membership',
+    )
+    def _check_hasfield(self, node):
+        # type: (astroid.Call) -> None
+        attr = node.func
+        if not isinstance(attr, astroid.Attribute) or attr.attrname not in ('HasField', 'ClearField'):
+            return
+        desc = _get_protobuf_descriptor(attr.expr)
+        if desc is None:
+            return
+        for arg in node.args:
+            for val in _get_inferred_values(arg):
+                if not hasattr(val, 'value'):
+                    continue
+                if val.value not in desc.field_names:
+                    self.add_message('protobuf-undefined-attribute', node=node, args=(val.value, desc.name))
+                    continue
+                fd = desc.fields_by_name[val.value]
+                if is_repeated(fd):
+                    self.add_message('protobuf-no-repeated-membership', node=node)
+                    continue
+                if desc.proto3:
+                    if not is_composite(fd) and not is_oneof(fd):
+                        # all fields in proto3 are labelled optional
+                        # "optional" fields are members of a singular oneof
+                        self.add_message('protobuf-no-proto3-membership', node=node, args=(val.value,))
+                        continue
+
     @utils.check_messages('protobuf-undefined-attribute')
     def visit_assignattr(self, node):
         # type: (astroid.AssignAttr) -> None
@@ -321,33 +399,38 @@ class ProtobufDescriptorChecker(BaseChecker):
             vals = node.expr.inferred()
         except astroid.InferenceError:
             return  # TODO: warn or redo
+        descriptors = []  # type: List[SimpleDescriptor]
         # Look for any version of the inferred type to be a Protobuf class
         for val in vals:
             cls_def = None
             if val is astroid.Uninferable:
-                continue
+                return  # break early (ref #44 and astroid 03d15b0)
             if hasattr(val, '_proxied'):
-                if wellknowntype(val):
-                    self._disable('no-member', node.lineno)
-                    return
                 cls_def = val._proxied  # type: astroid.ClassDef
-            elif isinstance(val, astroid.Module):
-                return self._check_module(val, node)  # FIXME: move
             elif isinstance(val, astroid.ClassDef):
                 cls_def = val
             if cls_def and getattr(cls_def, '_is_protobuf_class', False):
-                break  # getattr guards against Uninferable (always returns self)
+                # getattr guards against Uninferable (always returns self so can't use hasattr)
+                descriptors.append(cls_def._protobuf_descriptor)
+
+        found = None  # type: Optional[SimpleDescriptor]
+        missing = []  # type: List[SimpleDescriptor]
+        for desc in descriptors:
+            if node.attrname in desc.field_names or node.attrname in desc.extensions_by_name:
+                found = desc
+                self._disable('no-member', node.lineno)  # Should always be checked by us instead
+                break
+            missing.append(desc)
         else:
-            # couldn't find cls_def
-            return
-        desc = cls_def._protobuf_descriptor  # type: SimpleDescriptor
-        self._disable('no-member', node.lineno)  # Should always be checked by us instead
-        if node.attrname not in desc.field_names:
-            self.add_message('protobuf-undefined-attribute', args=(node.attrname, cls_def.name), node=node)
+            if len(missing) != 1:
+                return  # don't attempt to warn on multiple options
+            desc = missing[0]
+            self.add_message('protobuf-undefined-attribute', args=(node.attrname, desc.name), node=node)
+            self._disable('no-member', node.lineno)  # Should always be checked by us instead
             self._disable('assigning-non-slot', node.lineno)
-        else:
-            self._check_type_error(node, desc)
-            self._check_no_assign(node, desc)
+        if found is not None:
+            self._check_type_error(node, found)
+            self._check_no_assign(node, found)
 
     @utils.check_messages('protobuf-type-error')
     def _check_type_error(self, node, desc):
@@ -378,12 +461,30 @@ class ProtobufDescriptorChecker(BaseChecker):
         if is_composite(fd) or is_repeated(fd):
             self.add_message('protobuf-no-assignment', node=node, args=(desc.name, attr))
 
-    def _check_module(self, mod, node):
-        # type: (astroid.Module, astroid.Attribute) -> None
-        if not is_some_protobuf_module(mod):
+    def visit_subscript(self, node):
+        self._check_extension_getitem(node)
+
+    @utils.check_messages('protobuf-wrong-extension-scope')
+    def _check_extension_getitem(self, node):
+        # type: (astroid.Subscript) -> None
+        attr, slice = node.value, node.slice
+        if not isinstance(attr, astroid.Attribute) or attr.attrname != 'Extensions':
             return
-        if node.attrname not in mod.locals:
-            self.add_message('protobuf-undefined-attribute', args=(node.attrname, mod.name), node=node)
+        if not isinstance(slice, astroid.Index) or not isinstance(slice.value, astroid.Attribute):
+            return
+        target_desc = _get_protobuf_descriptor(attr.expr)
+        ext_desc = _get_protobuf_descriptor(slice.value.expr)
+        if target_desc is None:
+            return
+        ext_name = slice.value.attrname
+        if ext_desc is not None:
+            if ext_name not in ext_desc.extensions_by_name:
+                return  # should raise protobuf-undefined-attribute elsewhere
+            fd = ext_desc.extensions_by_name[ext_name]
+        else:  # ext_desc is None
+            return  # TODO for now
+        if not target_desc.is_extended_by(fd):
+            self.add_message('protobuf-wrong-extension-scope', node=node, args=(ext_name, target_desc.name))
 
     def _disable(self, msgid, line, scope='module'):
         try:
