@@ -1,3 +1,4 @@
+from keyword import iskeyword
 from functools import lru_cache
 from typing import Any, List, Tuple, Set, Dict, Union, MutableMapping, Iterator
 import textwrap
@@ -13,7 +14,8 @@ try:
 except ImportError:  # pragma: nocover
     import sys
     import warnings
-    if sys.version_info >= (3, 9):
+    from google.protobuf import __version__ as _protobuf_version
+    if _protobuf_version <= '3.15.0' and sys.version_info >= (3, 9):
         warnings.warn(
             "google.protobuf (earlier than 3.15.x) does not support Python 3.9"
             " (see https://github.com/protocolbuffers/protobuf/issues/7978)"
@@ -33,6 +35,12 @@ except ImportError:  # pragma: nocover
     class ScalarMap:
         pass
 
+try:
+    from google.protobuf.internal.well_known_types import WKTBASES
+except ImportError:
+    WKTBASES = {}
+
+
 PROTOBUF_IMPLICIT_ATTRS = [
     'ByteSize',
     'Clear',
@@ -41,6 +49,7 @@ PROTOBUF_IMPLICIT_ATTRS = [
     'CopyFrom',
     'DESCRIPTOR',
     'DiscardUnknownFields',
+    'Extensions',
     'HasExtension',
     'HasField',
     'IsInitialized',
@@ -68,12 +77,21 @@ def is_repeated(fd):
 
 def is_composite(fd):
     # type: (FieldDescriptor) -> bool
-    return fd.type == FieldDescriptor.TYPE_MESSAGE
+    return fd.type in (FieldDescriptor.TYPE_MESSAGE, FieldDescriptor.TYPE_GROUP)
 
 def is_map_field(fd):  # FIXME: too many selectors
     # type: (FieldDescriptor) -> bool
     return is_composite(fd) and fd.message_type.has_options and fd.message_type.GetOptions().map_entry
 
+def is_optional(fd):
+    # type: (FieldDescriptor) -> bool
+    # Only relevant for proto2
+    return fd.label == FieldDescriptor.LABEL_OPTIONAL
+
+def is_oneof(fd):
+    # type: (FieldDescriptor) -> bool
+    return fd.containing_oneof is not None
+
 
 class TODO(object):
     pass  # These fields are not assignable
@@ -117,6 +135,11 @@ def full_name(fd):
     return '.'.join(field_type_path(fd))
 
 
+def _nonprotected_members(cls):
+    # type: (type) -> Set[str]
+    return set(m for m in dir(cls) if not m.startswith('_'))
+
+
 class SimpleDescriptor(object):
     def __init__(self, desc):
         # type: (Union[EnumDescriptor, Descriptor]) -> None
@@ -125,8 +148,9 @@ class SimpleDescriptor(object):
             self._enum_desc = desc
         else:
             self._is_protobuf_enum = False
-            self._desc = desc
+            self._desc = desc  # type: Descriptor
         self._cls_hash = str(id(self))  # err...
+        self.bases = []
 
     def is_nested(self, fd):
         # type: (FieldDescriptor) -> bool
@@ -136,6 +160,15 @@ class SimpleDescriptor(object):
         # type: (FieldDescriptor) -> bool
         return fd.message_type is self._desc
 
+    def is_extended_by(self, fd):
+        # type: (FieldDescriptor) -> bool
+        return fd.is_extension and fd.containing_type is self._desc
+
+    @property
+    def proto3(self):
+        # type: () -> bool
+        return self._desc.file.syntax == 'proto3'
+
     @property
     def identifier(self):
         # type: () -> str
@@ -153,6 +186,14 @@ class SimpleDescriptor(object):
         else:
             return self._desc.name
 
+    @property
+    def full_name(self):
+        # type: () -> str
+        if self._is_protobuf_enum:
+            return self._enum_desc.full_name
+        else:
+            return self._desc.full_name
+
     @property
     def options(self):
         if self._desc.has_options:
@@ -163,6 +204,13 @@ class SimpleDescriptor(object):
                     return None
             return FalseyAttributes()
 
+    @property
+    def extensions_by_name(self):
+        # type: () -> Dict[str, FieldDescriptor]
+        if self.is_enum:
+            return dict()
+        return dict(self._desc.extensions_by_name)
+
     @property
     def fields(self):
         # type: () -> List[FieldDescriptor]
@@ -174,12 +222,15 @@ class SimpleDescriptor(object):
         if self._is_protobuf_enum:
             return set(self._enum_desc.values_by_name) | set(PROTOBUF_ENUM_IMPLICIT_ATTRS)
         else:
+            base_fields = set(PROTOBUF_IMPLICIT_ATTRS)  # TODO: move this into bases
+            for base_cls in self.bases:
+                base_fields |= _nonprotected_members(base_cls)
             desc = self._desc  # type: Descriptor
             return set(desc.fields_by_name) | \
                    set(desc.enum_values_by_name) | \
                    set(desc.enum_types_by_name) | \
                    set(desc.nested_types_by_name) | \
-                   set(PROTOBUF_IMPLICIT_ATTRS)
+                   base_fields
 
     @property
     def fields_by_name(self):
@@ -309,21 +360,37 @@ def _template_composite_field(parent_name, name, field_type, is_nested=False):
     """.format(name=name, field_type=field_type, qualifier=qualifier))
 
 
+def _to_module_name(fn):
+    # type: (str) -> str
+    """
+    Try to guess imported name from file descriptor path
+    """
+    fn = fn.replace('/', '_dot_')
+    fn = fn[:-len('.proto')]  # strip suffix
+    fn += '__pb2'  # XXX: might only be one underscore?
+    return fn
+
+
 def _template_message(desc, descriptor_registry):
     # type: (Descriptor, DescriptorRegistry) -> str
     """
     Returns cls_def string, list of fields, list of repeated fields
     """
+    this_file = desc.file
     desc = SimpleDescriptor(desc)
+    if desc.full_name in WKTBASES:
+        desc.bases.append(WKTBASES[desc.full_name])
     descriptor_registry[desc.identifier] = desc
 
     slots = desc.field_names
 
+    # TODO: refactor field partitioning and iskeyword checks
     # NOTE: the "pass" statement is a hack to provide a body when args is empty
     initialisers = ['pass']
     initialisers += [
         'self.{} = self.{}()  # inner_nonrepeated_fields'.format(field_name, field_type)
         for field_name, field_type in desc.inner_nonrepeated_fields
+        if not iskeyword(field_name)
     ]
 
 
@@ -331,6 +398,7 @@ def _template_message(desc, descriptor_registry):
     initialisers += [
         'self.{} = []  # repeated_fields'.format(field_name)
         for field_name in repeated_scalar_fields
+        if not iskeyword(field_name)
     ]
 
     rcfields = {
@@ -344,21 +412,45 @@ def _template_message(desc, descriptor_registry):
     initialisers += [
         _template_composite_field(desc.name, field_name, field_type, is_nested)
         for field_name, field_type, is_nested in repeated_composite_fields
+        if not iskeyword(field_name)
     ]
 
     # TODO: refactor this
     external_fields = [
-        (f.name, full_name(f.message_type)) for f in desc.message_fields
+        (f, f.message_type) for f in desc.message_fields
         if not desc.is_nested(f)
         if f not in rcfields  # don't want to double up above
     ]
+    siblings = [
+        (f, f.name, full_name(msg_type))
+        for f, msg_type in external_fields
+        if msg_type.file is this_file
+    ]
     initialisers += [
-        'self.{} = {}()  # external_fields'.format(field_name, field_type)
-        for field_name, field_type in external_fields
-
+        'self.{} = {}()  # external_fields (siblings)'.format(field_name, field_type)
+        for _, field_name, field_type in siblings
+        if not iskeyword(field_name)
+    ]
+    externals = [
+        (f, f.name, _to_module_name(msg_type.file.name), full_name(msg_type))  # TODO: look up name instead of heuristic?
+        for f, msg_type in external_fields
+        if msg_type.file is not this_file
+    ]
+    initialisers += [
+        'self.{} = {}.{}()  # external_fields (imports)'.format(field_name, qualifier, field_type)
+        for _, field_name, qualifier, field_type in externals
+        if not iskeyword(field_name)
     ]
 
-    args = ['self'] + ['{}=None'.format(f) for f in slots]
+    # Extensions should show up as attributes on message instances but not
+    # as keyword arguments in message constructors
+    initialisers += [
+        'self.{} = object()  # extensions'.format(ext_name)
+        for ext_name in desc.extensions_by_name
+        if not iskeyword(ext_name)
+    ]
+
+    args = ['self'] + ['{}=None'.format(f) for f in slots if not iskeyword(f)]
     init_str = 'def __init__({argspec}):\n{initialisers}\n'.format(
         argspec=', '.join(args),
         initialisers=textwrap.indent('\n'.join(initialisers), '    '),
@@ -454,6 +546,20 @@ def mod_node_to_class(mod, name):
     return ns[name]
 
 
+def resolve_imports(mod):
+    # type: (astroid.Module) -> List[str]
+    """
+    UNUSED: proposed to look up imported names rather than guess (see external fields clauses)
+    """
+    import_names = []
+    for node in mod.nodes_of_class((astroid.Import, astroid.ImportFrom)):
+        for original, alias in node.names:
+            name = alias or original
+            if name.endswith('_pb2'):
+                import_names.append(name)
+    return import_names
+
+
 def transform_module(mod):
     # type: (astroid.Module) -> astroid.Module
     for name in mod.wildcard_import_names():
@@ -470,4 +576,4 @@ def transform_module(mod):
 def is_some_protobuf_module(node):
     # type: (astroid.Module) -> bool
     modname = node.name
-    return modname.endswith('_pb2') and not modname.startswith('google.')
+    return modname.endswith('_pb2')
